#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdint.h>


#include "resource.h"

static int reading;

typedef struct buffer buffer_t;
#ifndef _BUFFER_PROTO
#define _BUFFER_PROTO 
#endif



_BUFFER_PROTO buffer_t *buffer_new(size_t blksize);
_BUFFER_PROTO void buffer_free(buffer_t *B, void **Rd, size_t *Rl);
_BUFFER_PROTO void buffer_collect(buffer_t *B, void **Rd, size_t *Rl);
_BUFFER_PROTO void buffer_write(buffer_t *B, void *d, size_t l);
_BUFFER_PROTO void buffer_append(buffer_t *dst, buffer_t *src_f);
_BUFFER_PROTO size_t buffer_read(buffer_t *B, void *d, size_t l);


struct buffer_node { 
  struct buffer_node *next;
  size_t len,size,start;
  unsigned char data[1];
};

struct buffer {
  struct buffer_node *first, *last;
  size_t blksize;
};



_BUFFER_PROTO buffer_t *buffer_new(size_t blksize)
{
  buffer_t *B;
  B= malloc(sizeof(*B));
  B->first= B->last= NULL;
  B->blksize= blksize;
  return B;
}



_BUFFER_PROTO void buffer_collect(buffer_t *B, void **Rd, size_t *Rl)
{
  size_t S;
  struct buffer_node *P;
  unsigned char *R;

  if (!Rd && !Rl) return ;

  S= 0;
  for(P=B->first;P;P=P->next) S+= P->len;

  if (S==0) { if (Rd) *Rd= NULL; if (Rl) *Rl= 0; return ; }
  if (!Rd) { *Rl= S; return ; }

  R= malloc(S);
  S= 0;
  for(P=B->first;P;P=P->next)
  {
    memcpy(R+S, P->data+P->start, P->len);
    S+= P->len;
  }

  *Rd= (void*) R;
  if (Rl) *Rl= S;
}

_BUFFER_PROTO void buffer_free(buffer_t *B, void **Rd, size_t *Rl)
{
  struct buffer_node *P,*N;
  if (Rd || Rl) buffer_collect(B, Rd, Rl);
  for(P=B->first;P;P=N)
  {
    N= P->next;
    free(P);
  }
  free(B);
}

_BUFFER_PROTO void buffer_write(buffer_t *B, void *d, size_t l)
{
  struct buffer_node *P;
  size_t cp;

  while(l)
  {
    P= B->last;
    if (!P || (P->start+P->len)==P->size)
    {
      P= malloc(sizeof(*P)+B->blksize-1);
      P->len= 0;
      P->next= NULL;
      P->size= B->blksize;
      P->start= 0;
      if (!B->last) { B->last= B->first= P; }
      else { B->last->next= P; B->last= P; }
    }

    cp= P->size-P->len-P->start;
    if (cp>l) cp= l; 

    memcpy(P->data+P->len+P->start, d, cp);
    P->len+= cp;
    l-= cp;
    d= (unsigned char*) d + cp;
  }
}

_BUFFER_PROTO void buffer_append(buffer_t *dst, buffer_t *src_f)
{
  if (src_f->first!=NULL)
  {
    if (dst->first==NULL)
    {
      dst->first= src_f->first;
      dst->last= src_f->last;
    }
    else
    {
      dst->last->next= src_f->first;
      dst->last= src_f->last;
    }
  }
  free(src_f);
}



_BUFFER_PROTO size_t buffer_read(buffer_t *B, void *d, size_t l)
{
  size_t total;
  struct buffer_node *P;
  size_t cp;
  total= 0;

  while(l)
  {
    P= B->first;
    if (!P) break;  
    if (P->len==0)
    {
      B->first= P->next;
      if (!B->first) B->last= NULL;
      free(P);
      continue;
    } 
    cp= P->len;
    if (cp>l) cp= l;
    memcpy(d, P->data+P->start, cp);
    P->start+= cp;
    P->len-= cp;
    total+= cp;
    d= (unsigned char*) d + cp;
  }
  return total;
}




typedef struct
{
  const char* name;
  void *arg;
  void (*func)(const char *, void *);
} field_t;

int nfields, sfields;
field_t *fields;
char *ifn;
FILE *fp;
char *ip;
char inbuf[2048];
int inbufl;
int lineno;

void die(const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  fprintf(stderr,"near %s:%d: ", ifn, lineno);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  fprintf(stderr,"\n");
  exit(1);
}


void inpline()
{
  if (!fgets(inbuf, sizeof(inbuf)-2, fp))
    die("unexpected end of file");
  lineno++;
  inbuf[sizeof(inbuf)-1]= 0;
  ip= inbuf;
}

void scan_past_char(int chr)
{
  while(1)
  {
    while(*ip) if (*ip++==chr) return ; 
    inpline();
  }
}

void skip_spaces()
{
again:
  while(isspace(*ip)) ip++;
  if (!*ip) { inpline(); goto again; }
}

char* get_fieldname()
{
  char *b,*e,*R;
  skip_spaces();
  if (*ip=='}') return NULL;
  if (*ip!='.') die("bad line");

  ip++;
  skip_spaces();
  b= ip;
  while(!isspace(*ip) && *ip!='=' && *ip!=0) ip++;

  if (*ip==0) die("bad line");
  e= ip;
  R= malloc(e-b+1); memcpy(R, b, e-b);  R[e-b]= 0;
  scan_past_char('='); 
  return R;
}

void field(void (*func)(const char*,void*),const char *name, void *arg)
{
  field_t *F;
  if (reading)
  {
    if (nfields==sfields)
    {
      sfields += sfields/2 + 20;
      fields= realloc(fields, sfields*sizeof(fields[0]));
    }
    F= fields + nfields++;
    F->name= name;
    F->arg= arg;
    F->func= func;
  }
  else
  {
    fprintf(fp, ".%s= ", name);
    (*func)(name, arg);
    fprintf(fp, ",\n");
  }
}

void init_read(char *fn)
{
  ip= inbuf; inbuf[0]= 0;
  reading= 1;
  field(NULL,NULL,NULL);
  fp= fopen(fn,"rb");
  ifn= fn;
  lineno= 0;
  if (!fp) die("can't open %s: %s",fn);
}

void begin_struct()
{
  if (reading)
  {
    field(NULL,NULL,NULL);
    scan_past_char('{'); // }
  }
  else
  {
    fprintf(fp, "{\n");
  }
}


void end_struct(int top)
{
  char *R;
  int i;
  if (!reading) { fprintf(fp,"}"); return ; }
again:
  skip_spaces();
  if (!(R=get_fieldname())) { ip++; if (!top) scan_past_char(','); return ; }

  for(i=nfields-1;fields[i].func;i--)
    if (!strcmp(fields[i].name, R))
    {
      (*fields[i].func)(fields[i].name,fields[i].arg);
      fields[i]= fields[nfields-1];
      nfields--;
      free(R);
      goto again;
    }
  die("unknown field %s",R);
}

unsigned long getint()
{
  unsigned long R;
  skip_spaces();
  R= strtoul(ip, &ip, 0);
  scan_past_char(','); 
  return R;
}

void io_color(const char *fname, void *arg)
{
  uint32_t *R= arg;
  if (reading) { *R= getint(); }
  else { if (*R<1000) fprintf(fp,"%d",*R);
                      else fprintf(fp,"0x%x", *R); }
}

void io_dim(const char *fname, void *arg) { io_color(fname,arg); }

void io_3dbox(const char *fname, void *arg)
{
  R_3DBOX *R= arg;
  begin_struct();
  field(io_color, "light", &R->light);
  field(io_color, "dark", &R->dark);
  field(io_color, "border", &R->border);
  field(io_color, "start", &R->start);
  field(io_color, "end", &R->start);
  end_struct(0);
}

void parse_grayimg(R_GRAYIMG *R)
{
  int i;
  char *ff;
  scan_past_char('{'); 

  R->w= R->h= -1;
  R->data= NULL;

again:
  ff= get_fieldname();
  if (!ff) goto parse_done;
  if (!strcmp(ff,"w")) { R->w= getint(); goto nextfield; }
  if (!strcmp(ff,"h")) { R->h= getint(); goto nextfield; }
  if (!strcmp(ff,"data"))
  {
    int i;
    if (R->h==-1 || R->w==-1) die("bad image specification");
    R->data=calloc(R->w*R->h,sizeof(R->data[0]));

    for(i=0;i<R->w*R->h;i++)
    {
      skip_spaces();
      if (*ip=='}') break;
      R->data[i]= getint();
    }
    scan_past_char('}');
    scan_past_char(',');
    goto nextfield;
  }
  die("unknown image field %s",ff);
nextfield:
  free(ff);
  goto again;
parse_done:
  if (!R->data || R->w==-1 || R->h==-1)
    die("bad image specification, all fields are mandatory");
  ip++;
  scan_past_char(',');
}

void output_grayimg(R_GRAYIMG *R)
{
  int i;
  fprintf(fp,"{ .w= %d, .h= %d, .data= (unsigned char[]) {", R->w, R->h);
  for(i=0;i<R->w*R->h;i++)
  {
    fprintf(fp,"%d,",R->data[i]);
    if (i%16==0) fprintf(fp,"\n");
  }
  fprintf(fp, "}, },\n");
}

void io_grayimg(const char *fname, void *arg)
{
  if (reading) parse_grayimg(arg);
          else output_grayimg(arg);
}


void parse_colorimg(R_COLORIMG *R)
{
  int i;
  char *ff;
  scan_past_char('{'); 

  R->w= R->h= -1;
  R->data= NULL;

again:
  ff= get_fieldname();
  if (!ff) goto parse_done;
  if (!strcmp(ff,"w")) { R->w= getint(); goto nextfield; }
  if (!strcmp(ff,"h")) { R->h= getint(); goto nextfield; }
  if (!strcmp(ff,"data"))
  {
    int i;
    if (R->h==-1 || R->w==-1) die("bad image specification");
    R->data=calloc(R->w*R->h,sizeof(R->data[0]));

    for(i=0;i<R->w*R->h;i++)
    {
      skip_spaces();
      if (*ip=='}') break;
      R->data[i]= getint();
    }
    scan_past_char('}');
    scan_past_char(',');
    goto nextfield;
  }
  die("unknown image field %s",ff);
nextfield:
  free(ff);
  goto again;
parse_done:
  if (!R->data || R->w==-1 || R->h==-1)
    die("bad image specification, all fields are mandatory");
  ip++;
  scan_past_char(',');
}

void output_colorimg(R_COLORIMG *R)
{
  int i;
  fprintf(fp,"{ .w= %d, .h= %d, .data= (unsigned int[]) {", R->w, R->h);
  for(i=0;i<R->w*R->h;i++)
  {
    if (R->data[i]<1000) fprintf(fp,"%d,",R->data[i]);
                    else fprintf(fp, "0x%x,",R->data[i]);
    if (i%16==0) fprintf(fp,"\n");
  }
  fprintf(fp, "}, },\n");
}

void io_colorimg(const char *fname, void *arg)
{
  if (reading) parse_colorimg(arg);
          else output_colorimg(arg);
}

void io_top(resource_t *R)
{
  begin_struct();
  field(io_3dbox, "button", &R->button);
  field(io_colorimg, "appicon", &R->appicon);
  end_struct(1);
}

resource_t *parse_resources(char *fn)
{
  resource_t *R; 
  init_read(fn);
  R= calloc(1,sizeof(*R));
  io_top(R);
  return R;
}

void output_resources(resource_t *R)
{
  reading= 0;
  io_top(R);
}

int main(int argc,char **argv)
{
  resource_t *R;
  R= parse_resources(argv[1]); 
  fp= stdout;
  output_resources(R);
  return 0;
}
